/*
 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2 of the License, or
 (at your option) any later version.
 
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.
 
 You should have received a copy of the GNU General Public License along
 with this program; if not, write to the Free Software Foundation, Inc.,
 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

#include "Config.h"

#include "wx/wx.h"
#include "wx/numdlg.h"

#include "OpenGL.h"
#include "main.h"
#include "LeakCheck.h"

#include "GUI/MeasureBar.h"
#include "GUI/GLPane.h"
#include "GUI/MainFrame.h"

#include "Midi/Sequence.h"
#include "Midi/Track.h"

#include <iostream>


namespace AriaMaestosa
{
	
  int insert_at_measure = -1;
  int remove_from = -1;
  int remove_to = -1;
	
// ****************************************************************************************************************
	
/*
 * The pop-up menu that is shown if you right-click on a selected measure in the measure bar
 */
	
class SelectedMenu : public wxMenu
{
	
	DECLARE_LEAK_CHECK();
	
public:
	
SelectedMenu() : wxMenu()
{
		INIT_LEAK_CHECK();
		
		Append(1,wxGetTranslation("Remove selected measures"));

}

/*
 * Removes selected measures
 */

void removeSelected(wxCommandEvent& event)
{
	// ask confirmation
	int answer = wxMessageBox( wxGetTranslation("Do you really want to remove selected measures?\nThis change cannot be undone."), wxT(""), wxYES_NO);
	if(answer == wxNO or answer == wxCANCEL) return;
	if(answer != wxYES and answer != wxOK)
	{
		std::cout << "Unknown answer from wxMessageBox in SelectedMenu::removeSelected() : " << answer << std::endl;
	}
	
	// tell all tracks to remove measures
	Sequence* seq = getMainFrame()->getCurrentSequence();
	const int trackAmount = seq->getTrackAmount();
	
	for(int n=0; n<trackAmount; n++)
	{
		seq->getTrack(n)->removeMeasures(remove_from, remove_to);
	}
	
	// shorten song accrodingly to the number of measures removed
	getMainFrame()->changeMeasureAmount( seq->getMeasureAmount() - (remove_to-remove_from) );
}

DECLARE_EVENT_TABLE();
};

BEGIN_EVENT_TABLE(SelectedMenu, wxMenu)
EVT_MENU(1,SelectedMenu::removeSelected)
END_EVENT_TABLE()

// ****************************************************************************************************************

/*
 * The pop-up menu that is shown if you right-click on an unselected measure in the measure bar
 */

class UnselectedMenu : public wxMenu
{
	
	DECLARE_LEAK_CHECK();
	
public:
	
	UnselectedMenu() : wxMenu()
	{
		INIT_LEAK_CHECK();
		
		Append(2,wxGetTranslation("Insert measures"));
		
	}
	
	void insert(wxCommandEvent& event)
	{
		int number = wxGetNumberFromUser( wxGetTranslation("Insert empty measures between measures ") + to_wxString(insert_at_measure+1) +
										  wxGetTranslation(" and ") + to_wxString(insert_at_measure+2),
										  wxGetTranslation("Amount: "), wxT(""), 4 /*default*/, 1 /*min*/);
		if(number==-1) return;
		
		Sequence* seq = getMainFrame()->getCurrentSequence();
		const int trackAmount = seq->getTrackAmount();
		for(int n=0; n<trackAmount; n++)
		{
			seq->getTrack(n)->insertEmptyMeasures(insert_at_measure+1, number);
		}
		
		getMainFrame()->changeMeasureAmount( seq->getMeasureAmount() + number );

	}
	
	DECLARE_EVENT_TABLE();
};

BEGIN_EVENT_TABLE(UnselectedMenu, wxMenu)
EVT_MENU(2,UnselectedMenu::insert)
END_EVENT_TABLE()

// ****************************************************************************************************************
	
	
MeasureBar::MeasureBar(GLPane* p)
{
	INIT_LEAK_CHECK();
	
	glPane = p;	
	mainFrame = glPane->mainFrame;
	somethingSelected = false;
	
	selectedMenu = new SelectedMenu();
	unselectedMenu = new UnselectedMenu();
	
	lastMeasureInDrag = -1;
}

MeasureBar::~MeasureBar()
{
	delete selectedMenu;
	delete unselectedMenu;
}

/*
 * Change the number of items in the selected vector sothat it contains the same amount of elements as the number of measures.
 */

void MeasureBar::updateVector(int newSize)
{
	while((int)selected.size() < newSize) selected.push_back(false);
	while((int)selected.size() > newSize) selected.erase(selected.begin()+selected.size()-1);
}

void MeasureBar::render(int measureBarY)
{
	// if measure amount changed and MeasureBar is out of sync with its current number of measures, fix it
	if((int)selected.size() != mainFrame->getCurrentSequence()->getMeasureAmount())
		updateVector(mainFrame->getCurrentSequence()->getMeasureAmount());
	
	glColor3f(1, 1, 0.9);
    glBegin(GL_QUADS);
    glVertex2f(0, measureBarY);
    glVertex2f(glPane->getWidth(), measureBarY);
    glVertex2f(glPane->getWidth(), measureBarY+20);
    glVertex2f(0, measureBarY+20);
    glEnd();
    
    
    // black line at the top and bottom
    glColor3f(0, 0, 0);
    glBegin(GL_LINES);
    glVertex2f(0, measureBarY+1);
    glVertex2f(glPane->getWidth(), measureBarY+1);
    
    glVertex2f(0, measureBarY+20);
    glVertex2f(glPane->getWidth(), measureBarY+20);
    glEnd();
    
    // vertical lines and mesure ID
    int measureID=0;
    glColor3f(0,0,0);
	
	const float x_initial = 90 - mainFrame->getCurrentSequence()->getXScrollInPixels();
	const float x_step = mainFrame->getCurrentSequence()->ticksPerMeasure()*mainFrame->getCurrentSequence()->getZoom();
	
    for(float n=x_initial; n<glPane->getWidth(); n += x_step)
	{
        
        measureID++;
        if(measureID > mainFrame->getCurrentSequence()->getMeasureAmount()) break;
        
		// if measure is selected, draw in blue
		if( selected[measureID-1] )
		{
			
			glColor3f(0.71, 0.84, 1);
			
			glBegin(GL_QUADS);
			glVertex2f(n, measureBarY+1);
			glVertex2f(n, measureBarY+19);
			glVertex2f(n+x_step, measureBarY+19);
			glVertex2f(n+x_step, measureBarY+1);
			glEnd();
			
			glColor3f(0,0,0);
		}
		
        // vertical line
        glBegin(GL_LINES);
        glVertex2f(n, measureBarY);
        glVertex2f(n, measureBarY+20);
        glEnd();
        
        // measure ID
        glRasterPos2f(n+5, measureBarY+15);
        char buffer[4];
        sprintf (buffer, "%d", measureID);
        
        for(int i=0; buffer[i]; i++)
		{
            glutBitmapCharacter(GLUT_BITMAP_HELVETICA_10, buffer[i]);
        }
    }
	
}

/*
 * When mouse button is pushed on Measure Bar
 */

void MeasureBar::mouseDown(int x, int y)
{

	const int measure_amount = selected.size();
	for(int n=0; n<measure_amount; n++)
	{
		selected[n] = false;
	}
	
	const float x1 = 90 - mainFrame->getCurrentSequence()->getXScrollInPixels();
	
	// length of a measure
	const float xstep = mainFrame->getCurrentSequence()->ticksPerMeasure()*mainFrame->getCurrentSequence()->getZoom();

	
	const int measure_vectorID = (int)( (x-x1)/xstep );
	
	//std::cout << "xstep=" << xstep << " x1=" << x1 << " x=" << x << " measure=" << (measure_vectorID+1) << std::endl;
	
	selected[measure_vectorID] = true;
	
	somethingSelected = true;
	lastMeasureInDrag = measure_vectorID;

	
}

/*
 * When mouse is held down and dragged on Measure Bar
 */

void MeasureBar::mouseDrag(int mousex_current, int mousey_current, int mousex_initial, int mousey_initial)
{
	const float x1 = 90 - mainFrame->getCurrentSequence()->getXScrollInPixels();
	const float xstep = mainFrame->getCurrentSequence()->ticksPerMeasure()*mainFrame->getCurrentSequence()->getZoom();

	const int measure_vectorID = (int)( (mousex_current-x1)/xstep );
	
	if(lastMeasureInDrag == -1) return; //invalid
	
	selected[measure_vectorID] = true;
	somethingSelected = true;
	
	// previous measure seelcted and this one and contiguous - select all other measures inbetween
	if( abs(lastMeasureInDrag-measure_vectorID) > 1 )
	{
		
		for(int n=measure_vectorID; n != lastMeasureInDrag; n += 
			(lastMeasureInDrag-measure_vectorID)/abs(lastMeasureInDrag-measure_vectorID))
		{
			selected[n] = true;
		}
	}
	
	lastMeasureInDrag = measure_vectorID;
}

/*
 * Selection is done, select notes contained in the selected measures
 */

void MeasureBar::mouseUp(int mousex_current, int mousey_current, int mousex_initial, int mousey_initial)
{
	if(lastMeasureInDrag == -1) return; //invalid
	
	Sequence* sequence = mainFrame->getCurrentSequence();
	
	const int measureAmount = selected.size();
	
	// determine selection range in midi ticks
	int minimal_tick = -1, maximal_tick = -1;
	for(int n=0; n<measureAmount; n++)
	{
		// iterate through measures to find the first selected one
		if(selected[n])
		{
			// we found a first selected measure, remember it as minimal tick
			minimal_tick = n * mainFrame->getCurrentSequence()->ticksPerMeasure();
			while( n++ and selected[n]) {} // skip all uneslected measures
			maximal_tick = n * mainFrame->getCurrentSequence()->ticksPerMeasure();
			break;
		}
	}
	
	// iterate through notes and select those that are within the selection range just found
	const int trackAmount = sequence->getTrackAmount();
	for(int track=0; track<trackAmount; track++)
	{
		const int noteAmount = sequence->getTrack(track)->getNoteAmount();
		for(int n=0; n<noteAmount; n++)
		{
			const int note_tick = sequence->getTrack(track)->getNoteStartInMidiTicks(n);
			// note is within selection range? if so, select it, else unselect it. 
			if( note_tick >= minimal_tick and note_tick < maximal_tick ) sequence->getTrack(track)->selectNote(n, true, true);
			else sequence->getTrack(track)->selectNote(n, false, true);
		}
	}
	
}

void MeasureBar::rightClick(int x, int y)
{
	const int x1 = 90 - mainFrame->getCurrentSequence()->getXScrollInPixels();
	const float xstep = mainFrame->getCurrentSequence()->ticksPerMeasure()*mainFrame->getCurrentSequence()->getZoom();
	const int measure_vectorID = (int)( (x-x1)/xstep );
	
	// is the clicked measure selected?
	if(selected[measure_vectorID])
	{
		const int measureAmount = selected.size();
		for(int n=0; n<measureAmount; n++)
		{
			// iterate through measures to find which measures are selected
			if(selected[n])
			{
				// we found a first selected measure, remember it
				remove_from = n;
				do{ n++; } while(selected[n]); // skip all selected measures
				remove_to = n;
				break;
			}
		}
		
		// unselect
		for(int n=remove_from; n<remove_to+1; n++)
		{
			selected[n]=false;
		}
		glPane->PopupMenu( (wxMenu*)selectedMenu, x, y);
	}
	else
	{
		// find on which measure the user clicked (substract by half-a-measure because we're inserting measures
		// between 2 measures so we need to know not n which measure user clicked but between which measures)
		insert_at_measure =  (int)(
									  (float)(x - x1 - xstep/2)/xstep
									  );

		glPane->PopupMenu( (wxMenu*)unselectedMenu, x, y);
	}

}

void MeasureBar::unselect()
{
	if(!somethingSelected) return;
	somethingSelected = false;
	
	const int measureAmount = selected.size();
	for(int n=0; n<measureAmount; n++)
	{
		selected[n] = false;
	}
	lastMeasureInDrag = -1;
}

}
